-------------------------------------------------------------------------------
-- Localization libraries
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("SimpleSelfRebuff")
local BS = AceLibrary("Babble-Spell-2.2")

-------------------------------------------------------------------------------
-- Aliases & constants
-------------------------------------------------------------------------------

local activationCheckEvents = {
	'PLAYER_DEAD', 'PLAYER_UNGHOST', 'PLAYER_ALIVE',
	'PLAYER_UPDATE_RESTING', 'PLAYER_AURAS_CHANGED',
	'PLAYER_CONTROL_GAINED', 'PLAYER_CONTROL_LOST',
}

local CATEGORY_TRACKING = L["Tracking"] or "Tracking"
local CATEGORY_MAINHAND = L["Main weapon"] or "Main weapon"
local CATEGORY_OFFHAND  = L["Off-hand weapon"] or "Off-hand weapon"

local SOURCE_SPELL = 'spell'
local SOURCE_ITEM  = 'item'

local TARGET_AURA     = 'aura'
local TARGET_TRACKING = 'tracking'
local TARGET_MAINHAND = 'main-hand'
local TARGET_OFFHAND  = 'off-hand'

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local categories = {}

local buffToCast
local castButton

local monitoringActive

local db

local sources = {}
local targets = {}

local loadFuncs = {}

-------------------------------------------------------------------------------
-- Externals
-------------------------------------------------------------------------------

local AceOO = AceLibrary('AceOO-2.0')
local Gratuity = AceLibrary('Gratuity-2.0')
local Waterfall = AceLibrary:HasInstance('Waterfall-1.0') and AceLibrary('Waterfall-1.0')
 
local Mixin = AceOO.Mixin
local Class = AceOO.Class
local Classpool = AceOO.Classpool

-------------------------------------------------------------------------------
-- Addon declaration
-------------------------------------------------------------------------------

SimpleSelfRebuff = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceModuleCore-2.0")
local SimpleSelfRebuff = SimpleSelfRebuff

-- Static early initialization
do
	local self = SimpleSelfRebuff

	self:RegisterDB("SimpleSelfRebuffDB")

	self.options = {	
		type = 'group',
		args = {
			menu = {
				type = 'execute',
				name = L["Menu"],
				desc = L["Menu"],
				func = function() Waterfall:Open('SimpleSelfRebuff') end,
				hidden = not Waterfall,
				wfHidden = true,
				order = 50,
			},
			menuHeader = {
				type = 'header',
				name = L["Global options"],
				order = 51,
				wfHidden = true,					
				hidden = not Waterfall,
			},		
			--
			rebuffThreshold = {
				type = 'range',
				name = L["Rebuff time threshold"],
				desc = L["Any buff with remaining time less than this value will be recast. Select 0 to disable."],
				min = 0,
				max = 180,
				step = 1,
				bigStep = 5,
				get = function() return db.rebuffThreshold end,
				set = function(value)
					db.rebuffThreshold = value
					self:CheckRebuff()
				end,
			},
			--
			buffHeader = {
				type = 'header',
				name = L["Buff selection"],
				order = 199,
				wfHidden = true,
			},
		}
	}
		
	self.CATEGORY_TRACKING = CATEGORY_TRACKING
	self.CATEGORY_MAINHAND = CATEGORY_MAINHAND 
	self.CATEGORY_OFFHAND  = CATEGORY_OFFHAND 
	
	self.SOURCE_SPELL = SOURCE_SPELL
	self.SOURCE_ITEM  = SOURCE_ITEM
	
	self.TARGET_AURA     = TARGET_AURA
	self.TARGET_TRACKING = TARGET_TRACKING 
	self.TARGET_MAINHAND = TARGET_MAINHAND
	self.TARGET_OFFHAND  = TARGET_OFFHAND

	-- For debugging
	SSR = self
	
	self.categories = categories
	self.sources = sources
	self.targets = targets
end

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function error(msg, ...)
	if select('#', ...) > 0 then
		msg = msg:format(...)
	end
	return _G.error(msg, 2)
end

local function erase(...)
	for i=1,select('#', ...) do
		local t = select(i, ...)
		if type(t) == 'table' then
			for k in pairs(t) do
				t[k] = nil 
			end
		end
	end
end

local function loadHash(t, ...)
	for i = 1, select('#', ...), 2 do
		local k,v = select(i, ...)
		if t[k] == nil then
			t[k] = v
		end
	end
end

local function colorize(text, r, g, b)
	return ("|cff%02x%02x%02x%s|r"):format(math.floor(r*255), math.floor(g*255), math.floor(b*255), text)
end
SimpleSelfRebuff.colorize = colorize

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == 'table' then
			setmetatable(t, nil)
			for k in pairs(t) do
				t[k] = nil
			end
			list[t] = true
		end
		return nil
	end
end

local function formatDuration(duration, default)
	if not duration or duration < 1 then
		return default or "-"
	elseif duration >= 3600 then
		return HOUR_ONELETTER_ABBR:format(math.ceil(duration/3600))
	elseif duration >= 60 then
		return MINUTE_ONELETTER_ABBR:format(math.ceil(duration/60))
	else
		return SECOND_ONELETTER_ABBR:format(duration)
	end
end
SimpleSelfRebuff.formatDuration = formatDuration

-------------------------------------------------------------------------------
-- Initializing
-------------------------------------------------------------------------------

function SimpleSelfRebuff:_rev(tag)
	self.revision = math.max(self.revision or 0, tonumber(tag:match("%d+")))
end

-- Addon setup
function SimpleSelfRebuff:OnInitialize()
	self:_rev("$Revision: 43075 $")
	self.version = "r" .. SimpleSelfRebuff.revision
	self.date = string.sub("$Date: 2007-07-07 09:51:48 -0400 (Sat, 07 Jul 2007) $", 8, 17)

	self:RegisterDefaults('profile', {
		categories = {},
		rebuffThreshold = 30,
	})
		
	db = self.db.profile

	self:RegisterChatCommand({"/ssr", "/simpleselfrebuff"}, self.options)
	
	local AceDebug = AceLibrary:HasInstance('AceDebug-2.0') and AceLibrary('AceDebug-2.0')
	if AceDebug then
		AceDebug:embed(self)
		local opt = AceDebug:GetAceOptionsDataTable(self)
		self.options.args.debug = opt.debug
		--SimpleSelfRebuff:SetDebugging(UnitName('player') == 'Adirelle')
	else
		function self.Debug() end
	end
	self.modulePrototype.Debug = function(module, ...) self:Debug(...) end
	
	if Waterfall then
		Waterfall:Register('SimpleSelfRebuff', 
			'aceOptions', self.options,
			'title','SimpleSelfRebuff',
			'colorR',0.8,
			'colorG',0.2,
			'colorB',0.8
		)
	end
		
end

function SimpleSelfRebuff:OnEnable(first)
	if first then
		self:ProcessRegistry()
	end

	self:RegisterBucketEvent(activationCheckEvents, 0, 'CheckActivation')
	self:RegisterBucketEvent('SimpleSelfRebuff_BuffChanged', 0, 'CheckRebuff')
	self:RegisterEvent('SimpleSelfRebuff_AvailableBuffsChanged', 'AvailableBuffsChanged')

	self:CheckActivation()
end

function SimpleSelfRebuff:OnDisable(first)
	self:SetMonitoringActive(false)
end

function SimpleSelfRebuff:AvailableBuffsChanged()
	self:UpdateOptions()
	self:CheckActivation()
end

-------------------------------------------------------------------------------
-- Module modulePrototype
-------------------------------------------------------------------------------

SimpleSelfRebuff:SetModuleMixins('AceEvent-2.0')
SimpleSelfRebuff.modulePrototype.core = SimpleSelfRebuff

function SimpleSelfRebuff:OnModuleCreate(name, module)
	self:SetModuleDefaultState(module, true)
end

function SimpleSelfRebuff.modulePrototype:OnInitialize()
	self.db = self.core:AcquireDBNamespace(self.name)
	
	if type(self.OnModuleInitialize) == "function" then
		self:OnModuleInitialize()
	end
	
	local options = self.options or {}
	
	options.enable = {
		type = 'toggle',
		name = L['Enable'],
		desc = L['Enable %s.'];format(self.optionName or self.name),
		order = 1,
		handler = self.core,
		passValue = self,
		get = "IsModuleActive",
		set = "ToggleModuleActive",
	}
	
	self.core.options.args[self.name] = {
		type = 'group',
		name = self.optionName or self.name,
		desc = self.optionDsc or self.name,
		handler = self,
		args = options,
		order = self.optionOrder,
	}	
	
end

function SimpleSelfRebuff.modulePrototype:OnEnable(first)
	if type(self.OnModuleEnable) == "function" then
		self:OnModuleEnable(first)
	end
end

function SimpleSelfRebuff.modulePrototype:OnDisable()
	if type(self.OnModuleDisable) == "function" then
		self:OnModuleDisable()
	end
end

function SimpleSelfRebuff.modulePrototype:RegisterDefaults(kind, defaults)
	self.core:RegisterDefaults(self.name, kind, defaults)
end

-------------------------------------------------------------------------------
-- Options handling
-------------------------------------------------------------------------------

do
	local NO_BUFF = L["None"]
	
	local function getSelectCategoryBuff(category)
		local buff = category:GetExpectedBuff()
		return buff and buff.name or NO_BUFF
	end
	
	local function getToggleCategoryBuff(category)
		local buff = category:GetExpectedBuff()
		return buff and (buff.name == category.name)
	end
	
	local function setCategoryBuff(category, name)
		if name == NO_BUFF then
			name = nil
		elseif name == true then
			name = category.name
		end
		category:SetExpectedBuff(name)
	end
	
	function SimpleSelfRebuff:BuildOptions()
		local opts = self.options.args
		for categoryName, category in pairs(categories) do
			local categoryName, category = categoryName, category
			local option = {
				name = categoryName,
				desc = categoryName,
				passValue = category,
				set = setCategoryBuff,
				order = 200,
			}
			if category.count > 1 then
				option.type = 'text'
				option.get = getSelectCategoryBuff
				local validate = { NO_BUFF }
				option.validate = validate
				option.disabled = function() return #validate < 2 end
			else
				option.type = 'toggle'
				option.get = getToggleCategoryBuff
				local theBuff = select(2, next(category.buffs))
				option.disabled = function() return not theBuff:IsCastable() end
			end
			opts[categoryName] = option
		end
		self:UpdateOptions()
	end
	
	function SimpleSelfRebuff:UpdateOptions()
		for categoryName, category in pairs(categories) do
			if category.count > 1 then
				local validate = self.options.args[category.name].validate
				erase(validate)
				for name,buff in pairs(category.buffs) do
					if buff:IsCastable() then
						tinsert(validate, buff.name)
					end
				end
				table.sort(validate)				
				tinsert(validate, 1, NO_BUFF)
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Activation
-------------------------------------------------------------------------------

function SimpleSelfRebuff:SetMonitoringActive(value)
	if value and not monitoringActive then
		monitoringActive = true
		self:Debug('Activated')
		
		self:ScheduleRepeatingEvent('SimpleSelfRebuff_BuffChanged', 1.5)
		
		self:TriggerEvent('SimpleSelfRebuff_MonitoringEnabled')
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
		
	elseif not value and monitoringActive then
		monitoringActive = nil
		self:Debug('Disactivated')

		self:CancelScheduledEvent('SimpleSelfRebuff_BuffChanged')
		
		self:TriggerEvent('SimpleSelfRebuff_MonitoringDisabled')		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

function SimpleSelfRebuff:CheckActivation()
	self:SetMonitoringActive(
		not (UnitIsDeadOrGhost('player') or IsResting())
	)
end

function SimpleSelfRebuff:IsMonitoringActive()
	return monitoringActive
end

-------------------------------------------------------------------------------
-- Core
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckRebuff()

	local rebuffThreshold = db.rebuffThreshold > 0 and db.rebuffThreshold or nil
	local newBuffToCast, buffTimeLeft, shortestBuff, shortestBuffDuration
	
	for categoryName,category in pairs(categories) do
		local expectedBuff = category:GetExpectedBuff()
		if expectedBuff and expectedBuff:IsUsable() then			
			local actualBuff, _, timeLeft = category:GetActualBuff()
			
			-- Check if it is missing
			if not actualBuff then
				newBuffToCast = expectedBuff
				--self:Debug('Missing buff: %q', expectedBuff)
				break
				
			-- Check rebuffing threshold
			elseif rebuffThreshold and timeLeft and actualBuff == expectedBuff then
				if not shortestBuffDuration or timeLeft < shortestBuffDuration then
					shortestBuffDuration, shortestBuff = timeLeft, actualBuff
				end
			end
				
		end
	end
	
	if not newBuffToCast and shortestBuff and shortestBuffDuration < rebuffThreshold then
		newBuffToCast = shortestBuff
		buffTimeLeft = shortestBuffDuration
		self:Debug('Buff needs refresh: %q (%q)', newBuffToCast, shortestBuffDuration)
	end

	if newBuffToCast ~= buffToCast then
		buffToCast = newBuffToCast
		self:TriggerEvent('SimpleSelfRebuff_PendingCastChanged', buffToCast, buffTimeLeft)
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end

end

function SimpleSelfRebuff:GetPendingCast()
	return buffToCast
end

-------------------------------------------------------------------------------
-- Base buff class
-------------------------------------------------------------------------------
local CategoryClass, BuffClass

BuffClass = Class()

function BuffClass.prototype:init(name, category, ...)
	if type(name) ~= 'string' then
		error('Argument #2 to BuffClass:new() should be a string, not %q', type(name))
	end		
	
	self.name = name
	self.category = category
	loadHash(self, ...)
	
	BuffClass.super.prototype.init(self)
	
	local targetMixin = targets[self.target]
	targetMixin.buffs[self.name] = self
	targetMixin.allBuffs[self] = true
	
	local sourceMixin = sources[self.source]
	sourceMixin.buffs[self.name] = self
	sourceMixin.allBuffs[self] = true
end

function BuffClass.prototype:IsCastable()
	if self.found then
		if self.minLevel and UnitLevel('player') < self.minLevel then
			return false
		elseif self.maxLevel and UnitLevel('player') > self.maxLevel then
			return false
		elseif type(self.IsTargetCastable) == "function" then
			return self:IsTargetCastable()
		end
		return true
	end
	return false
end

function BuffClass.prototype:SetAsActualBuff(flag, timeLeft)
	if flag then
		self.category:SetActualBuff(self, timeLeft)
	else
		self.category:SetActualBuff(nil, nil)
	end
end

function BuffClass.prototype:IsActualBuff()
	return self.category:GetActualBuff() == self
end

function BuffClass.prototype:IsExpectedBuff()
	return self.category:GetExpectedBuff() == self
end

-------------------------------------------------------------------------------
-- Category class
-------------------------------------------------------------------------------

CategoryClass = Class()

function CategoryClass.prototype:init(name, target, source, ...)
	if type(name) ~= 'string' then
		error('Argument #2 to CategoryClass:new() should be a string, not %q', type(name))
	end
	CategoryClass.super.prototype.init(self)
	self.name = name
	self.buffs = {}
	self.count = 0
	self.target = target or TARGET_AURA
	self.source = source or SOURCE_SPELL
	loadHash(self, ...)
end

function CategoryClass.prototype:add(name, ...)

	if self.buffs[name] then
		error("%q already registered in category %q", name, self.name)
	end
	
	local source = self.source
	local target = self.target
	
	if not sources[source] and loadFuncs[source] then
		loadFuncs[source]()
		loadFuncs[source] = nil
	end
	if not sources[source] then
		error("unknown source for %q: %q", name, source)
	end		
	
	if not targets[target] and loadFuncs[target] then
		loadFuncs[target]()
		loadFuncs[target] = nil
	end	
	if not targets[target] then
		error("unknown target for %q: %q", name, target)
	end		
		
	local class = Classpool(BuffClass, sources[source], targets[target])
	local buff = class:new(name, self, ...)

	self.buffs[buff.name] = buff
	self.count = self.count + 1

	return self
end

function CategoryClass.prototype:addMulti(...)
	for i=1,select('#', ...) do
		local name = select(i, ...)
		self:add(name)
	end
	return self
end

function CategoryClass.prototype:GetExpectedBuff()
	local name = db.categories[self.name]
	if self.buffs[name] then
		return self.buffs[name]
	else
		db.categories[self.name] = nil
	end
end

function CategoryClass.prototype:SetExpectedBuff(buff)
	local buffName 
	if type(buff) == 'string' then
		buffName = buff
	elseif buff then
		buffName = buff.name
	end
	if buffName and not self.buffs[buffName] then
		error("Unknown buff %q in category %q", buffName, self.name)
	end	
	db.categories[self.name] = buffName
	SimpleSelfRebuff:TriggerEvent('SimpleSelfRebuff_BuffSetupChanged')
end

function CategoryClass.prototype:GetActualBuff()
	local timeLeft = self.actualBuffExpiration and math.floor(self.actualBuffExpiration - GetTime())
	if not timeLeft or timeLeft <= 0 then
		timeLeft = nil
	end
	return self.actualBuff, timeLeft
end

function CategoryClass.prototype:SetActualBuff(buff, timeLeft)
	local expected = self:GetExpectedBuff()
	if expected and expected.subcat and expected.subcat ~= buff.subcat then
		return
	end	
	if not buff or not timeLeft or timeLeft == 0 then
		timeLeft = nil
	end
	local expiration = timeLeft and GetTime() + timeLeft or nil
	local dirty = buff ~= self.actualBuff or expiration ~= self.actualBuffExpiration
	self.actualBuff = buff
	self.actualBuffExpiration = expiration
	if dirty then
		SimpleSelfRebuff:TriggerEvent('SimpleSelfRebuff_BuffChanged')
		return true
	end
end

-------------------------------------------------------------------------------
-- Some mixin helpers
-------------------------------------------------------------------------------

local BuffMixin, BuffSourceMixin, BuffTargetMixin

do
	
	function BuffMixin(...)
		local mixin = Mixin{...}
		AceLibrary('AceEvent-2.0'):embed(mixin)		
		
		mixin.allBuffs = {}
		mixin.buffs = {}
		
		mixin.Debug = function(mixin, ...) SimpleSelfRebuff:Debug(...) end
		
		function mixin:_OnMixinEnable()
			if not self._enabled then
				self:Debug('Enabling %q', mixin)
				if type(self.OnEnable) == "function" then
					self:OnEnable()
				end		
				self._enabled = true
			end
		end
		
		function mixin:_OnMixinDisable()
			if self._enabled then
				self:Debug('Disabling %q', mixin)
				self:UnregisterAllEvents()
				if type(self.OnDisable) == "function" then
					self:OnDisable()
				end
				self:_RegisterMixinEvents()
			end
		end
		
		function mixin:_RegisterMixinEvents()
			self:RegisterEvent('SimpleSelfRebuff_MonitoringEnabled', '_OnMixinEnable')
			self:RegisterEvent('SimpleSelfRebuff_MonitoringDisabled', '_OnMixinDisable')
		end
		
		mixin:_RegisterMixinEvents()
		
		return mixin
	end
	
	function BuffSourceMixin(source, ...)
		mixin = BuffMixin("IsUsable", "SetupBuffSource", "source", ...)
		mixin.source = source
		
		function mixin:ToString()
			return 'BuffSource-'..self.source
		end
		
		sources[source] = mixin
		return mixin
	end
	
	function BuffTargetMixin(target, ...)
		mixin = BuffMixin("IsTargetCastable", "SetupBuffTarget", "target", ...)
		mixin.target = target
		mixin.IsTargetCastable = returnTrue

		function mixin:ToString()
			return 'BuffTarget-'..self.target
		end

		targets[target] = mixin
		return mixin
	end
	
end

-------------------------------------------------------------------------------
-- Buffs from items
-------------------------------------------------------------------------------

loadFuncs[SOURCE_ITEM] = function()
	SimpleSelfRebuff:Debug("Loading buff-from-item support")
	
	local ItemBuff = BuffSourceMixin(SOURCE_ITEM, 'ResolveItemId')
	local missingItems
		
	function ItemBuff:OnInstanceInit(buff)
		if not buff.itemid then
			error("itemid is missing for ItemBuff %q", buff.name)
		end
		buff:ResolveItemId()
	end
	
	function ItemBuff:OnEnable()
		self:ResolveMissingItems()
	end
	
	function ItemBuff:ResolveMissingItems()
		if not missingItems then
			return
		end
		self:Debug('Scanning missing items')
		
		for buff in pairs(missingItems) do
			local oldName = buff.name
			local found, linked = buff:ResolveItemId(true) 
			if found and buff.name ~= oldName then
				self.buffs[oldName] = nil
				self.buffs[buff.name] = buff
			end
			if linked then
				break
			end
		end
		
		if missingItems then
			self:Debug('Planning scanning event')
			self:ScheduleEvent('SimpleSelfRebuff_ResolveItemId', self.ResolveMissingItems, 0.5, self)
		end
	end
	
	function ItemBuff:ResolveItemId(setLink)
		local itemName = GetItemInfo(self.itemid)
		if itemName then
			
			if missingItems and missingItems[self] then
				missingItems[self] = nil
				if not next(missingItems) then
					missingItems = del(missingItems)
				end
			end
			
			self.found = true
			
			if itemName ~= self.name then
				ItemBuff:Debug('%s solved as %s', self.name, itemName)
				local buffs = self.category.buffs
				buffs[self.name] = nil
				buffs[itemName] = self
				self.name = itemName
			end
			
			return true
			
		elseif setLink then
			ItemBuff:Debug('Linking #%d', self.itemid)			
			Gratuity:SetHyperLink('item:'..self.itemid)
			return false, true
			
		else
			ItemBuff:Debug('Register #%d for scanning', self.itemid)			
			if not missingItems then
				missingItems = new()
			end
			missingItems[self] = true			
		end
	end
	
	function ItemBuff:IsUsable()
		local start, duration = GetItemCooldown(self.itemid)
		if start > 0 and duration > 0 and start + duration > GetTime() then
			return false
		end					
		return (IsUsableItem(self.itemid))
	end
	
	function ItemBuff:SetupBuffSource(castButton)
		self:Debug('Setup using for item %q', self.name)
		castButton:SetAttribute('type', 'item')
		castButton:SetAttribute('item', self.name)
	end
	
end

-------------------------------------------------------------------------------
-- Buffs from spells
-------------------------------------------------------------------------------

loadFuncs[SOURCE_SPELL] = function()
	SimpleSelfRebuff:Debug('Loading buff-from-spell support')
		
	local SpellBuff = BuffSourceMixin(SOURCE_SPELL)
		
	function SpellBuff:OnEnable()
		if next(self.allBuffs) then
			self:RegisterEvent('SPELLS_CHANGED', 'ScanSpells')
			self:ScanSpells()
		end
	end
	
	function SpellBuff:ScanSpells()
		local dirty = false
		local knownSpells = new()
		
		local i = 1
		while true do
			local name = GetSpellName(i, BOOKTYPE_SPELL)
			if not name then
				break
			end
			knownSpells[name] = true
			i = i + 1
		end
		
		for buff in pairs(self.allBuffs) do
			local found = knownSpells[buff.name]
			if found and not buff.found then
				-- SimpleSelfRebuff:Debug('New spell: %q', buff.name)
				buff.found = true
				dirty = true
			elseif not found and buff.found then
				-- SimpleSelfRebuff:Debug('Old spell: %q', buff.name)
				buff.found = nil
				dirty = true
			end
		end
		
		knownSpells = del(knownSpells)
	
		if dirty then
			self:TriggerEvent('SimpleSelfRebuff_AvailableBuffsChanged')
		end	
	end
	
	function SpellBuff:IsUsable()
		local start, duration = GetSpellCooldown(self.name)
		if start > 0 and duration > 0 and start + duration > GetTime() then
			return false
		end					
		return (IsUsableSpell(self.name))
	end
	
	function SpellBuff:SetupBuffSource(castButton)
		SimpleSelfRebuff:Debug('Setup casting for spell %q', self.name)
		castButton:SetAttribute('type', 'spell')
		castButton:SetAttribute('spell', self.name)
	end
	
end

-------------------------------------------------------------------------------
-- Auras
-------------------------------------------------------------------------------

loadFuncs[TARGET_AURA] = function()
	SimpleSelfRebuff:Debug('Loading aura buff support')

	local AuraBuff = BuffTargetMixin(TARGET_AURA)
	
	function AuraBuff:OnEnable()
		if next(self.allBuffs) then
			self:RegisterEvent('PLAYER_AURAS_CHANGED', 'ScanAuras')
			self:RegisterEvent('SimpleSelfRebuff_BuffSetupChanged', 'ScanAuras')
			self:ScanAuras()
		end
	end
	
	function AuraBuff:ScanAuras()
		local hasTimedBuff
		local seen = new()
		
		for i = 1, 32 do
			-- Only take in account buffs we actually cast
			local buffName, _, _, _, duration, timeLeft = UnitBuff('player', i)	
			if not buffName then
				break
			end
			local buff = buffName and self.buffs[buffName]
			if duration and buff then
				if timeLeft > 0 then
					hasTimedBuff = true
				end
				seen[buff] = true
				if buff:SetAsActualBuff(true, timeLeft) then
					self:Debug('Found aura %s (%q)', buff.name, timeLeft)
				end
			end
		end
		
		for buff in pairs(self.allBuffs) do
			if not seen[buff] and buff:IsActualBuff() then
				self:Debug('%s faded out', buff.name)
				buff:SetAsActualBuff(false)
			end
		end
		
		seen = del(seen)
		
		if hasTimedBuff then
			self:ScheduleEvent('SimpleSelfRebuff_ScanAuras', self.ScanAuras, 5, self)
		end
	end
	
	function AuraBuff:SetupBuffTarget(castButton)
		castButton:SetAttribute('unit', 'player')
	end
	
end

-------------------------------------------------------------------------------
-- Tracking
-------------------------------------------------------------------------------

loadFuncs[TARGET_TRACKING] = function()
	SimpleSelfRebuff:Debug('Loading tracking support')
	
	local TrackingBuff = BuffTargetMixin(TARGET_TRACKING)
	TrackingBuff.textures = {}
	
	function TrackingBuff:OnInstanceInit(buff)
		local texture = BS:GetSpellIcon(buff.name)
		if not texture then
			error('Unknown tracking texture for spell %q', buff.name)
		elseif self.textures[texture] then
			error('Tracking texture already set for %q: %q', buff.name, self.textures[texture].name)
		end
		self.textures[texture] = buff
	end
	
	function TrackingBuff:OnEnable()
		self:RegisterEvent('PLAYER_AURAS_CHANGED', 'ScanTracking')
		self:RegisterEvent('SimpleSelfRebuff_BuffSetupChanged', 'ScanTracking')
		self:ScanTracking()
	end
	
	function TrackingBuff:ScanTracking()
		local texture = GetTrackingTexture()
		local buff = texture and self.textures[texture]
		if buff then
			buff:SetAsActualBuff(true)
		else
			for buff in pairs(self.allBuffs) do
				if buff:IsActualBuff() then
					buff:SetAsActualBuff(false)
				end
			end
		end
	end
	
	function TrackingBuff:SetupBuffTarget(castButton)
		castButton:SetAttribute('unit', 'player')
	end

end

-------------------------------------------------------------------------------
-- Weapon enchantment buff
-------------------------------------------------------------------------------

-- Thoses are pretty ugly to handle due to poor API
loadFuncs[TARGET_MAINHAND] = function()
	SimpleSelfRebuff:Debug('Loading weapon buff support')

	local isValidItemEquipLoc = {
		[16] = {
			INVTYPE_WEAPON         = true,
			INVTYPE_2HWEAPON       = true,
			INVTYPE_WEAPONMAINHAND = true,
		},
		[17] = {
			INVTYPE_WEAPON         = true,
			INVTYPE_WEAPONOFFHAND  = true,
		}
	}
	
	local function BuildWeaponBuffMixin(target, slotName, infoIndex)
		local slot = GetInventorySlotInfo(slotName)
		if type(slot) ~= 'number' then
			error("Invalid slot name %q => %q", slotName, slot)
		end
		
		local mixin = BuffTargetMixin(target, "slot")
		
		mixin.slot = slot
		mixin.infoIndex = infoIndex
		
		function mixin:OnEnable()
			if next(self.allBuffs) then
				self:RegisterBucketEvent('UNIT_INVENTORY_CHANGED', 0.5, 'CheckSlot')
				self:ScheduleRepeatingEvent('SimpleSelfRebuff_ScanWeaponBuffs', self.ScanWeaponBuffs, 0.5, self)
				self:RegisterEvent('SimpleSelfRebuff_BuffSetupChanged', 'ScanWeaponBuffs')
				self:CheckSlot()
				self:ScanWeaponBuffs()
			end
		end
		
		function mixin:ScanWeaponBuffs()
			local hasEnchant, timeLeft = select(self.infoIndex, GetWeaponEnchantInfo())
			
			if hasEnchant and not self.hasEnchant then
				self.hasEnchant = true
				local aBuff = next(self.allBuffs)
				local category = aBuff.category
				category:GetExpectedBuff():SetAsActualBuff(true, timeLeft/1000)				
				
			elseif not hasEnchant and self.hasEnchant then
				self.hasEnchant = nil
				for buff in pairs(self.allBuffs) do
					if buff:IsActualBuff() then
						buff:SetAsActualBuff(false)
					end
				end
				
			end			
		end
		
		function mixin:CheckSlot()
			local itemLink = GetInventoryItemLink("player", self.slot)
			local itemEquipLoc = itemLink and select(9, GetItemInfo(itemLink))
			if itemEquipLoc ~= self.itemEquipLoc then
				self.itemEquipLoc = itemEquipLoc
				self:TriggerEvent('SimpleSelfRebuff_AvailableBuffsChanged')
			end
		end		
		
		-- Embedded methods
		function mixin:SetupBuffTarget(castButton)
			SimpleSelfRebuff:Debug('Setting up for buff %q on %q', self.name, self.slot)
			castButton:SetAttribute('target-slot', self.slot)
		end
		
		function mixin:IsTargetCastable()
			return mixin.itemEquipLoc and isValidItemEquipLoc[mixin.slot][mixin.itemEquipLoc]
		end

	end
	
	BuildWeaponBuffMixin(TARGET_MAINHAND, 'MainHandSlot', 1)
	BuildWeaponBuffMixin(TARGET_OFFHAND, 'SecondaryHandSlot', 4)
	
end

loadFuncs[TARGET_OFFHAND] = loadFuncs[TARGET_MAINHAND]

-------------------------------------------------------------------------------
-- Buff registry
-------------------------------------------------------------------------------

do
	local setupFuncs = {}

	function SimpleSelfRebuff:RegisterBuffSetup(func)
		if type(func) == "function" then
			tinsert(setupFuncs, func)
		else
			error("Arg #1 to RegisterBuffs must be a function, not %q", type(func))
		end
	end
	
	function SimpleSelfRebuff:ProcessRegistry()
		if not setupFuncs then
			return
		end
		
		-- Prebuild categories
		self:GetCategory(self.CATEGORY_TRACKING, self.TARGET_TRACKING)
		self:GetCategory(self.CATEGORY_MAINHAND, self.TARGET_MAINHAND)
		self:GetCategory(self.CATEGORY_OFFHAND,  self.TARGET_OFFHAND)
		
		-- Load all funcs
		local selfClassName = select(2, UnitClass('player'))
		for i,func in ipairs(setupFuncs) do
			local succes, msg = pcall(func, self, BS, selfClassName, L)
			if not succes then
				geterrorhandler()(("Error in setupFunc #%d: %s"):format(i, msg))
			end
		end
		
		-- Clean up
		setupFuncs = del(setupFuncs)
		loadFuncs = del(loadFuncs)
		
		self:BuildOptions()
	end
end
	
function SimpleSelfRebuff:GetCategory(name, ...)
	if type(name) ~= 'string' then
		error("Argument #2 to GetCategory should be a string, not %q", type(name))
	end
	if not categories[name] then
		categories[name] = CategoryClass:new(name, ...)
	end
	return categories[name]
end

function SimpleSelfRebuff:AddStandaloneBuff(name, ...)
	if type(name) ~= 'string' then
		error("Argument #2 to AddStandaloneBuff should be a string, not %q", type(name))
	end
	self:GetCategory(name):add(name, ...)
end

function SimpleSelfRebuff:AddMultiStandaloneBuffs(...)
	for i=1,select('#', ...) do
		local name = select(i, ...)
		self:AddStandaloneBuff(name)
	end
end
